#pragma once

#include <iostream>
#include <string>

//#define TEST
//#define PRINT
#define PET 0
#define MR 1
#define APPLICATION PET

#define VERIFY_CONDITION( CONDITION, ERROR_MESSAGE ) { verifyCondition(CONDITION, ERROR_MESSAGE); }

inline void verifyCondition(bool condition, std::string errorMessage)
{
	if (!condition)
	{
		std::cerr << "Error: " << errorMessage << std::endl;
		exit(EXIT_FAILURE);
	}
}

struct vec3 {
	float x, y, z;
	vec3() { x = y = z = 0; }
	vec3(float _x, float _y, float _z) { x = _x; y = _y; z = _z; }
	vec3 operator/(float r) { return vec3(x / r, y / r, z / r); }
	vec3 operator/(const vec3& r) { return vec3(x / r.x, y / r.y, z / r.z); }
	vec3 operator*(const vec3& r) { return vec3(x * r.x, y * r.y, z * r.z); }
	vec3 operator*(float r) { return vec3(x * r, y * r, z * r); }
	void operator*=(float r) { x *= r; y *= r; z *= r; }
	vec3 operator+(const vec3& r) { return vec3(x + r.x, y + r.y, z + r.z); }
	vec3 operator-(const vec3& r) { return vec3(x - r.x, y - r.y, z - r.z); }
	bool operator==(const vec3& r) {
//		printf("%f, %f, %f <-> %f, %f, %f\n", x, y, z, r.x, r.y, r.z);
		return (x == r.x && y == r.y && z == r.z); 
	}
};

inline vec3 vmax(const vec3& l, const vec3& r) {
	return vec3(fmax(l.x, r.x), fmax(l.y, r.y), fmax(l.z, r.z));
}
inline vec3 vmin(const vec3& l, const vec3& r) {
	return vec3(fmin(l.x, r.x), fmin(l.y, r.y), fmin(l.z, r.z));
}
inline float dot(const vec3& l, const vec3& r) {
	return (l.x * r.x + l.y * r.y + l.z * r.z);
}
inline vec3 cross(const vec3& v1, const vec3& v2) {
	return vec3(v1.y * v2.z - v1.z * v2.y, v1.z * v2.x - v1.x * v2.z, v1.x * v2.y - v1.y * v2.x);
}

struct ivec3 {
	int x, y, z;
	ivec3() { x = y = z = 0; }
	ivec3(int _x, int _y, int _z) { x = _x; y = _y; z = _z; }
	ivec3 operator*(const ivec3& r) { return ivec3(x * r.x, y * r.y, z * r.z); }
};
