#include "GLGraphics.h"

#include "glm/gtc/matrix_transform.hpp"

Texture::Texture() {
	width = height = 0;
	reset();
}

Texture::Texture(int width, int height, const std::vector<float>& image) {
	this->width = width;
	this->height = height;

	glGenTextures(1, &textureId);  				// id generation
	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_FLOAT, &image[0]); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Init()
{
	glGenTextures(1, &textureId);  				// id generation
}

void Texture::Load(int width, int height, const float* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_FLOAT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load(int width, int height, const unsigned char* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_UNSIGNED_BYTE, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load(int width, int height, const unsigned short* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_UNSIGNED_SHORT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load3(int width, int height, const float* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGB, GL_FLOAT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load3(int width, int height, const unsigned char* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load3(int width, int height, const unsigned short* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGB, GL_UNSIGNED_SHORT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load4(int width, int height, const float* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGBA, GL_FLOAT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load4(int width, int height, const unsigned char* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::Load4(int width, int height, const unsigned short* image)
{
	this->width = width;
	this->height = height;

	glBindTexture(GL_TEXTURE_2D, textureId);    // binding

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGBA, GL_UNSIGNED_SHORT, image); // To GPU
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void Texture::SetUniform(unsigned shaderProg, const char * samplerName, unsigned int textureUnit) {
	int location = glGetUniformLocation(shaderProg, samplerName);
	if (location >= 0) {
		glUniform1i(location, textureUnit);
		glActiveTexture(GL_TEXTURE0 + textureUnit);
		glBindTexture(GL_TEXTURE_2D, textureId);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void Texture::reset()
{
}

void GPUProgram::getErrorInfo(unsigned int handle) { // shader error report
	int logLen, written;
	glGetShaderiv(handle, GL_INFO_LOG_LENGTH, &logLen);
	if (logLen > 0) {
		char * log = new char[logLen];
		glGetShaderInfoLog(handle, logLen, &written, log);
		printf("Shader log:\n%s", log);
		delete[] log;
	}
}

void GPUProgram::checkShader(unsigned int shader, const char * message) { 	// check if shader could be compiled
	int OK;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &OK);
	if (!OK) { printf("%s!\n", message); getErrorInfo(shader); getchar(); }
}

void GPUProgram::checkLinking(unsigned int program) { 	// check if shader could be linked
	int OK;
	glGetProgramiv(program, GL_LINK_STATUS, &OK);
	if (!OK) { printf("Failed to link shader program!\n"); getErrorInfo(program); getchar(); }
}

GPUProgram::GPUProgram() { shaderProgramId = 0; }

unsigned int GPUProgram::getId() { return shaderProgramId; }

void GPUProgram::Create(const char * const vertexSource, const char * const fragmentSource, const char * const fragmentShaderOutputName) {
	// Create vertex shader from string
	unsigned int vertexShader = glCreateShader(GL_VERTEX_SHADER);
	if (!vertexShader) {
		printf("Error in vertex shader creation\n");
		exit(1);
	}
	glShaderSource(vertexShader, 1, &vertexSource, NULL);
	glCompileShader(vertexShader);
	checkShader(vertexShader, "Vertex shader error");

	// Create fragment shader from string
	unsigned int fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	if (!fragmentShader) {
		printf("Error in fragment shader creation\n");
		exit(1);
	}

	glShaderSource(fragmentShader, 1, &fragmentSource, NULL);
	glCompileShader(fragmentShader);
	checkShader(fragmentShader, "Fragment shader error");

	shaderProgramId = glCreateProgram();
	if (!shaderProgramId) {
		printf("Error in shader program creation\n");
		exit(1);
	}
	glAttachShader(shaderProgramId, vertexShader);
	glAttachShader(shaderProgramId, fragmentShader);

	// Connect the fragmentColor to the frame buffer memory
	glBindFragDataLocation(shaderProgramId, 0, fragmentShaderOutputName);	// this output goes to the frame buffer memory

																			// program packaging
	glLinkProgram(shaderProgramId);
	checkLinking(shaderProgramId);

	// make this program run
	glUseProgram(shaderProgramId);
}

void GPUProgram::Use() { 		// make this program run
	glUseProgram(shaderProgramId);
}

void GPUProgram::SetUniform1f(const char * samplerName, float f) {
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform1f(location, f);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniform1i(const char * samplerName, int i) {
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform1i(location, i);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniform2i(const char * samplerName, int i1, int i2) {
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform2i(location, i1, i2);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniform2f(const char * samplerName, float f1, float f2) {
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform2f(location, f1, f2);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniform3f(const char * samplerName, float f1, float f2, float f3) {
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform3f(location, f1, f2, f3);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniform4f(const char* samplerName, float f1, float f2, float f3, float f4)
{
	int location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniform4f(location, f1, f2, f3, f4);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void GPUProgram::SetUniformMat4(const char* samplerName, const glm::mat4& m)
{
	GLuint location = glGetUniformLocation(shaderProgramId, samplerName);
	if (location >= 0) {
		glUniformMatrix4fv(location, 1, GL_FALSE, &m[0][0]);
	}
	else printf("Uniform %s cannot be set\n", samplerName);
}

GPUProgram::~GPUProgram() { glDeleteProgram(shaderProgramId); }

GLfloat Quad::vertices[18] = { -1.0f, 1.0f, 0.0f,
1.0f, 1.0f, 0.0f,
-1.0f, -1.0f, 0.0f,
-1.0f, -1.0f, 0.0f,
1.0f, 1.0f, 0.0f,
1.0f, -1.0f, 0.0f };

GLfloat Quad::texCoords[12] = { 0.0f, 1.0f,
1.0f, 1.0f,
0.0f, 0.0f,
0.0f, 0.0f,
1.0f, 1.0f,
1.0f, 0.0f };

Quad::Quad()
{
}

Quad::~Quad()
{
	glDeleteBuffers(1, &vertexBuffer);
	glDeleteVertexArrays(1, &vertexArray);
	glDeleteVertexArrays(1, &texCoordBuffer);
}

void Quad::init()
{
	glGenVertexArrays(1, &vertexArray);
	glBindVertexArray(vertexArray);

	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 18, vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer((GLuint)0, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glGenBuffers(1, &texCoordBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, texCoordBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 12, texCoords, GL_STATIC_DRAW);
	glEnableVertexAttribArray(2);
	glVertexAttribPointer((GLuint)2, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindVertexArray(0);
}

void Quad::render() {
	glBindVertexArray(vertexArray);
	glDrawArrays(GL_TRIANGLES, 0, 6);
	glBindVertexArray(0);
}

void Quad::render(int width, int height)
{
	int vp[4];
	glGetIntegerv(GL_VIEWPORT, vp);
	glViewport(0, 0, width, height);
	glBindVertexArray(vertexArray);
	glDrawArrays(GL_TRIANGLES, 0, 6);
	glBindVertexArray(0);
	glViewport(vp[0], vp[1], vp[2], vp[3]);
}

Texture3D::~Texture3D()
{
	Reset();
}

void Texture3D::Init()
{
	glGenTextures(1, &textureId);
}

void Texture3D::Load(int width, int height, int depth, const glm::vec3& voxSize, const float* image)
{
	this->width = width;
	this->height = height;
	this->depth = depth;
	this->voxSize = voxSize;

	glBindTexture(GL_TEXTURE_3D, textureId);    // binding
	glTexImage3D(GL_TEXTURE_3D, 0, GL_R32F, width, height, depth, 0, GL_RED, GL_FLOAT, image); // To GPU
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // sampling
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	glBindTexture(GL_TEXTURE_3D, 0);
}

void Texture3D::SetUniform(unsigned shaderProg, const char* samplerName, unsigned int textureUnit)
{
	int location = glGetUniformLocation(shaderProg, samplerName);
	if (location >= 0) {
		glUniform1i(location, textureUnit);
		glActiveTexture(GL_TEXTURE0 + textureUnit);
		glBindTexture(GL_TEXTURE_3D, textureId);
	}
	else printf("uniform %s cannot be set\n", samplerName);
}

void Texture3D::SetFiltering(GLenum name, GLint value)
{
	glBindTexture(GL_TEXTURE_3D, textureId);
	glTexParameteri(GL_TEXTURE_3D, name, value);
	glBindTexture(GL_TEXTURE_3D, 0);
}

void Texture3D::Reset()
{
	glDeleteBuffers(1, &textureId);
	textureId = 0;
	width = height = depth = 0;
}

void OrbitingCamera::Scroll(float scroll)
{
	distance *= scroll;
}

void OrbitingCamera::Rotate(float x, float y)
{
	float dir = 1.0f;
	if (invert) dir = -1.0f;
	horizontalAngle += dir * x;
	verticalAngle = glm::clamp(verticalAngle + dir * y, -89.0f, 89.9f);
}

glm::mat4 OrbitingCamera::GetViewMat() const
{
	glm::vec3 pos = GetPosition();
	glm::mat4 V = glm::lookAt(pos, glm::vec3(0), glm::vec3(0, 0, 1));

	return V;
}

glm::vec3 OrbitingCamera::GetPosition() const
{
	glm::vec3 direction(
		cos(glm::radians(verticalAngle)) * sin(glm::radians(horizontalAngle)),
		cos(glm::radians(verticalAngle)) * cos(glm::radians(horizontalAngle)),
		sin(glm::radians(verticalAngle))
	);

	return direction * distance;
}