#pragma once

#include <vector>

#include "GL/gl3w.h"
#include "GL/glcorearb.h"
#include "glm/glm.hpp"

struct Texture {
	unsigned int textureId;
	int width, height;

	Texture();
	Texture(int width, int height, const std::vector<float>& image);
	void Init();
	void Load(int width, int height, const float* image);
	void Load(int width, int height, const unsigned char* image);
	void Load(int width, int height, const unsigned short* image);
	void Load3(int width, int height, const float* image);
	void Load3(int width, int height, const unsigned char* image);
	void Load3(int width, int height, const unsigned short* image);
	void Load4(int width, int height, const float* image);
	void Load4(int width, int height, const unsigned char* image);
	void Load4(int width, int height, const unsigned short* image);
	void SetUniform(unsigned shaderProg, const char * samplerName, unsigned int textureUnit = 0);
	void reset();
};


class Texture3D {
public:
	GLuint textureId{ 0 };
	int width{ 0 };
	int height{ 0 };
	int depth{ 0 };
	glm::vec3 voxSize;

	Texture3D() = default;
	~Texture3D();

	void Init();
	void Load(int width, int height, int depth, const glm::vec3& voxSize, const float* image);
	void SetUniform(unsigned shaderProg, const char * samplerName, unsigned int textureUnit = 0);
	void SetFiltering(GLenum name, GLint value);
	void Reset();
};

class GPUProgram {
	unsigned int shaderProgramId;

	void getErrorInfo(unsigned int handle);
	void checkShader(unsigned int shader, const char * message);
	void checkLinking(unsigned int program);
public:
	GPUProgram();

	unsigned int getId();
	void Create(const char * const vertexSource, const char * const fragmentSource, const char * const fragmentShaderOutputName);
	void Use();
	void SetUniform1f(const char * samplerName, float f = 1);
	void SetUniform1i(const char * samplerName, int i = 1);
	void SetUniform2i(const char * samplerName, int i1 = 1, int i2 = 1);
	void SetUniform2f(const char * samplerName, float f1 = 1, float f2 = 1);
	void SetUniform3f(const char * samplerName, float f1 = 1, float f2 = 1, float f3 = 1);
	void SetUniform4f(const char * samplerName, float f1 = 1, float f2 = 1, float f3 = 1, float f4 = 1);
	void SetUniformMat4(const char* samplerName, const glm::mat4& m);
	~GPUProgram();
};

class Quad {
private:
	GLuint vertexArray;

	static GLfloat vertices[18];
	GLuint vertexBuffer;

	static GLfloat texCoords[12];
	GLuint texCoordBuffer;

public:
	Quad();
	~Quad();

	void init();
	void render();
	void render(int widtgh, int height);
};

class OrbitingCamera
{
private:
	float horizontalAngle{ 180.0f };
	float verticalAngle{ 45.0f };
	float distance{ 5.0f };
	bool invert{ true };

public:
	OrbitingCamera() = default;
	~OrbitingCamera() = default;

	void Scroll(float scroll);
	void Rotate(float x, float y);
	glm::mat4 GetViewMat() const;
	glm::vec3 GetPosition() const;
	void SetInvert(bool invert) { this->invert = invert; }
	bool GetInvert() const { return invert; }
};
