#pragma once

#include <array>


namespace ColorMap
{
	constexpr size_t Size = 256;
	constexpr size_t SizeRGB = Size * 3;

	std::array<float, SizeRGB> grayScale;
	std::array<float, SizeRGB> inverseGrayScale;
	std::array<float, SizeRGB> NIH; // from Amide
    std::array<float, SizeRGB> inverseNIH; // from Amide

    std::array<float, 3> hsv_to_rgb(std::array<float, 3> hsv, float RGBMAX = 1.0f)
    {
        std::array<float, 3> rgb;
        float fraction, p, q, t;
        int whole;

        if (hsv[1] == 0.0f)
            rgb[0] = rgb[1] = rgb[2] = RGBMAX * hsv[2];
        else {
            if (hsv[0] == 360.0f)
                hsv[0] = 0.0f; /* from [0 360] to [0 360) */

            whole = (int)floor(hsv[0] / 60.0f);
            fraction = (hsv[0] / 60.0f) - whole;

            p = hsv[2] * (1.0f - hsv[1]);
            q = hsv[2] * (1.0f - (hsv[1] * fraction));
            t = hsv[2] * (1.0f - (hsv[1] * (1.0f - fraction)));

            switch (whole) {
            case 0:
                rgb[0] = RGBMAX * hsv[2];
                rgb[1] = RGBMAX * t;
                rgb[2] = RGBMAX * p;
                break;
            case 1:
                rgb[0] = RGBMAX * q;
                rgb[1] = RGBMAX * hsv[2];
                rgb[2] = RGBMAX * p;
                break;
            case 2:
                rgb[0] = RGBMAX * p;
                rgb[1] = RGBMAX * hsv[2];
                rgb[2] = RGBMAX * t;
                break;
            case 3:
                rgb[0] = RGBMAX * p;
                rgb[1] = RGBMAX * q;
                rgb[2] = RGBMAX * hsv[2];
                break;
            case 4:
                rgb[0] = RGBMAX * t;
                rgb[1] = RGBMAX * p;
                rgb[2] = RGBMAX * hsv[2];
                break;
            case 5:
                rgb[0] = RGBMAX * hsv[2];
                rgb[1] = RGBMAX * p;
                rgb[2] = RGBMAX * q;
                break;
            default:
                /* should never get here */
                rgb = { 0, 0, 0 };
                break;
            }
        }

        return rgb;
    }

    std::array<float, 3> NIH_color(float v)
    {
        float hue = 0.0f;
        float saturation = 1.0f;
        float value = 1.0f;
        if (v > 0.2f)
            value = 1.0;
        else
            value = 5.0f * v;

        hue = 300.0f * (1.0f - v);

        return hsv_to_rgb({ hue, saturation, value });
    }

	void Init()
	{
		for (size_t i = 0; i < Size; ++i)
		{
			size_t iRed   = 3 * i + 0;
			size_t iGreen = 3 * i + 1;
			size_t iBlue  = 3 * i + 2;
			float v = static_cast<float>(i) / Size;
			// Grayscale
			grayScale[iRed] = grayScale[iGreen] = grayScale[iBlue] = v;
			// Inverse Grayscale
			inverseGrayScale[iRed] = inverseGrayScale[iGreen] = inverseGrayScale[iBlue] = 1.0f - v;
			// NIH
            auto rgb = NIH_color(v);
			NIH[iRed] = rgb[0];
			NIH[iGreen] = rgb[1];
			NIH[iBlue] = rgb[2];
            // inverseNIH
            rgb = NIH_color(1-v);
            inverseNIH[iRed] = rgb[0];
            inverseNIH[iGreen] = rgb[1];
            inverseNIH[iBlue] = rgb[2];
		}
	}
}

