#pragma once

const char * const vertexSource = R"(
	#version 130
	precision highp float;

	//layout(location = 0) in vec3 vp;
	out vec3 fp;

	void main() {
		//gl_Position = vec4(vp, 1.0);
		gl_Position = vec4(gl_Vertex.xyz, 1.0);
		fp = (gl_Vertex.xyz + 1.0) / 2.0;
	}
)";

// fragment shader in GLSL
const char * const fragmentSource = R"(
	#version 130
	precision highp float;

	uniform sampler3D inputImage;
	uniform float inputScaler;
	uniform int inputMode;
	uniform vec2 inputMinMax;
	uniform ivec2 windowSize;
	uniform vec2 inputImageScaler;
	uniform ivec2 inputCenter;
	uniform int flipVertical, flipHorizontal;
	uniform float inputImageSlice;
	uniform int imageAxis;
	uniform int sourceAxis;

	in vec3 fp;
	out vec4 outColor;

	vec4 getColor(sampler3D t, vec3 p, int m, ivec2 center, float slice, vec2 scaler)
	{
		ivec2 imageAxes = ivec2(0, 1);
		if (imageAxis != sourceAxis) {
			switch (imageAxis)
			{
			case 0:
				imageAxes = (sourceAxis == 1) ? ivec2(2, 1) : ivec2(1, 2);
				break;
			case 1:
				imageAxes = (sourceAxis == 2) ? ivec2(0, 2) : ivec2(2, 1);
				break;
			case 2:
				imageAxes = (sourceAxis == 0) ? ivec2(2, 0) : ivec2(0, 2);
				break;
			}
		}

		ivec3 imageSize = textureSize(t, 0);
		vec2 coordinateScaler = vec2(float(imageSize[imageAxes.x]) / float(windowSize.x),
									 float(imageSize[imageAxes.y]) / float(windowSize.y));

		vec2 cPos = vec2(p.x, 1.0 - p.y);
		cPos = cPos / coordinateScaler * scaler;
		cPos = cPos - vec2(float(center.x) / float(imageSize[imageAxes.x]), float(center.y) / float(imageSize[imageAxes.y])) * scaler;

		if(cPos.x < 0.0 || cPos.x > 1.0 || cPos.y < 0.0 || cPos.y > 1.0)
		{
			return vec4(0.85f, 0.8f, 0.71f, 1.0f);
		}

		// Flip vertically or horizontally
		cPos = vec2(flipHorizontal == 1 ? 1 - cPos.x : cPos.x, flipVertical == 1 ? 1.0 - cPos.y : cPos.y);

		vec3 samplePos = vec3(cPos, slice);
		switch(imageAxis)
		{
		case 0:
			if (sourceAxis == 1) samplePos = vec3(slice, cPos.y, cPos.x);
			else 
            if (sourceAxis == 2) samplePos = vec3(slice, cPos.x, 1 - cPos.y);
			break;
		case 1:
			if (sourceAxis == 0) samplePos = vec3(slice, cPos.y, cPos.x);
			else 
            if (sourceAxis == 2) samplePos = vec3(cPos.x, slice, 1 - cPos.y);
			break;
		case 2:
			if (sourceAxis == 0) samplePos = vec3(cPos.y, slice, cPos.x);
			else 
            if (sourceAxis == 1) samplePos = vec3(cPos.x, slice, cPos.y);
			break;
		}

		vec4 c = texture(t, samplePos);
		switch(m)
		{
		case 0:
			return vec4(c.xyz, 1.0);
			break;
		case 1:
			return vec4(vec3(c.x), 1.0);
			break;
		}
		return vec4(1.0, 0.0, 0.0, 1.0);
	}

	void main() {
		// switch(displayMode) ... was removed from the skeleton

		outColor = getColor(inputImage, fp, inputMode, inputCenter, inputImageSlice, inputImageScaler) * inputScaler;
	}
)";

const char * const vertexSourceVolumeRender = R"(
	#version 130
	precision highp float;

	//layout(location = 0) in vec3 vp;
	out vec3 fp;

	void main() {
		//gl_Position = vec4(vp, 1.0);
		gl_Position = vec4(gl_Vertex.xyz, 1.0);
		fp = (gl_Vertex.xyz + 1.0) / 2.0;
	}
)";

const char* const fragmentSourceVolumeRender = R"(
	#version 130
	precision highp float;

	#define vclamp 0.05

	uniform int renderMode;
	uniform float quality;
	uniform float aspect; // assumption: height / width
	uniform vec2 volumeAspect; // (height / width, depth / width)
	uniform vec4 volumeResolution;
	uniform float densityScale;
	uniform float densityNormalize;
	uniform mat4x4 invViewMat;
	uniform vec4 bgColor;
	uniform vec4 oLightDir;
	uniform vec4 lightColor;
	uniform float lightingStrength;
	uniform float isoValue;
	uniform float alphaScale;
	uniform float borderAlphaScale;
	uniform float borderWidth;
	uniform int shade;
	uniform float invert;
	uniform int useCullPlane;
	uniform vec4 cullPlane;
	uniform vec3 uvMin;
	uniform vec3 uvMax;

	uniform sampler3D textureData;
	uniform sampler2D colorMap;
	uniform sampler2D alphaMap;

	in vec3 fp;
	out vec4 outColor;

	struct Ray {
		vec3 origin;
		vec3 direction;
	};

	struct AABB {
		vec3 boxMin;
		vec3 boxMax;
	};

	// intersect ray with a box
	// http://www.siggraph.org/education/materials/HyperGraph/raytrace/rtinter3.htm
	bool intersectBox(Ray r, AABB aabb, out float tnear, out float tfar)
	{
		// compute intersection of ray with all six bbox planes
		vec3 invR = vec3(1.0f, 1.0f, 1.0f) / r.direction;
		vec3 tbot = invR * (aabb.boxMin - r.origin);
		vec3 ttop = invR * (aabb.boxMax - r.origin);

		// re-order intersections to find smallest and largest on each axis
		vec3 tmin = min(ttop, tbot);
		vec3 tmax = max(ttop, tbot);

		// find the largest tmin and the smallest tmax
		float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
		float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

		tnear = largest_tmin;
		tfar = smallest_tmax;

		return smallest_tmax > largest_tmin;
	}

	float restrictCullPlanes(vec3 uv)
	{
		return (useCullPlane > 0) ? (dot(vec4(uv, 1), cullPlane) >= 0 ? 1 : 0) : 1;
	}

	float restrictSlice(vec3 uv)
	{
		return step(uvMin.x, uv.x) * step(uvMin.y, uv.y) * step(uvMin.z, uv.z) * step(uv.x, uvMax.x) * step(uv.y, uvMax.y) * step(uv.z, uvMax.z)
			* restrictCullPlanes(uv);
	}

	float getDensity_restricted(sampler3D t, vec3 uv, float scale)
	{
		float d = texture(t, uv).r * densityNormalize * scale;
		return d * restrictSlice(uv);
	}

	float getDensity(sampler3D t, vec3 uv) 
	{
		float d = texture(t, uv).r * densityNormalize;
		return d;
	}

	vec3 getNormal(sampler3D t, vec3 uv)
	{
		vec3 normal;

		float dx = 1.0 / volumeResolution.x;
		float dy = 1.0 / volumeResolution.y;
		float dz = 1.0 / volumeResolution.z;
		normal.x = getDensity(t, uv + vec3(dx, 0, 0)) - getDensity(t, uv - vec3(dx, 0, 0));
		normal.y = getDensity(t, uv + vec3(0, dy, 0)) - getDensity(t, uv - vec3(0, dy, 0));
		normal.z = getDensity(t, uv + vec3(0, 0, dz)) - getDensity(t, uv - vec3(0, 0, dz));

		if (dot(normal, normal) < 0.001) {
			normal = vec3(0, 0, 1);
		}

		return normalize(normal);
	}

	float densityToAlpha(float density)
	{
		return texture(alphaMap, vec2(density, 0.5)).r * alphaScale;
	}

	vec3 densityToColor(float d)
	{
		//return vec3(d, d*d, d*d*d);
		return texture(colorMap, vec2(d, 0.5)).rgb;
	}

	float borderDist(vec3 uv)
	{
		float dx = min(abs(uvMin.x - uv.x), abs(uvMax.x - uv.x));
		float dy = min(abs(uvMin.y - uv.y), abs(uvMax.y - uv.y));
		float dz = min(abs(uvMin.z - uv.z), abs(uvMax.z - uv.z));
		return clamp(min(dx, min(dy, dz)) / borderWidth, 0, 1);
	}

	vec4 alphaBlend(vec4 bg, float maxStep, vec3 step, vec3 pos)
	{
		vec4 sum;
		vec3 lightDir = normalize(oLightDir.xyz);
		for (int i = 0; i < maxStep; ++i) {
			vec3 uv = pos / vec3(1,volumeAspect) + vec3(0.5, 0.5, 0.5);
			float density = getDensity_restricted(textureData, uv, 1);

			vec3 color = densityToColor(density * densityScale);
			if (shade != 0)
			{
				vec3 normal = getNormal(textureData, uv).xyz;
				float k = (1.0 - lightingStrength) + lightingStrength * dot(normalize(lightDir), normal);
				color.xyz *= k * lightColor.xyz;
			}
			float alpha = densityToAlpha(density);
			alpha *= mix(borderAlphaScale, 1, borderDist(uv));
			sum = (1.0 - sum.a) * vec4(color * alpha, alpha) + sum;

			pos += step;
		}
		//const float c = 0.5;
		//if (sum.a < c) return mix(bg, sum, smoothstep(0, 1, sum.a/c));
		//return sum;
		return mix(bg, sum, smoothstep(0, 1, sum.a));
	}

	vec4 xray(vec4 bg, float maxStep, vec3 step, vec3 pos)
	{
		float sum;
		for (int i = 0; i < maxStep; ++i) {
			vec3 uv = pos / vec3(1,volumeAspect) + vec3(0.5, 0.5, 0.5);
			float density = getDensity_restricted(textureData, uv, densityScale);

			sum += density;

			pos += step;
		}
		sum /= maxStep;
		sum = clamp(sum,0,1);
		sum = abs(invert-sum);
		if (invert > 0)
		{
			if (sum > 1-vclamp) return mix(bg, vec4(sum, sum, sum, 1), smoothstep(0, 1, (1-sum)/vclamp));
			return vec4(sum, sum, sum, 1);
		}
		if (sum < vclamp) return mix(bg, vec4(sum, sum, sum, 1), smoothstep(0, 1, sum/vclamp));
		return vec4(sum, sum, sum, 1);
	}

	vec4 mip(vec4 bg, float maxStep, vec3 step, vec3 pos)
	{
		float maxVal = 0;
		for (int i = 0; i < maxStep; ++i) {
			vec3 uv = pos / vec3(1,volumeAspect) + vec3(0.5, 0.5, 0.5);
			float density = getDensity_restricted(textureData, uv, densityScale);

			maxVal = max(density, maxVal);

			pos += step;
		}
		
		maxVal = clamp(maxVal,0,1);
		maxVal = abs(invert-maxVal);
		/*if (invert > 0)
		{
			if (maxVal > 1-vclamp) return mix(bg, vec4(maxVal, maxVal, maxVal, 1), smoothstep(0, 1, (1-maxVal)/vclamp));
			return vec4(maxVal, maxVal, maxVal, 1);
		}*/
		vec3 color = densityToColor(maxVal);
		if (maxVal < vclamp) return mix(bg, vec4(color, 1), smoothstep(0, 1, maxVal/vclamp));
		return vec4(color, 1);
	}

	vec4 isoSurface(vec4 bg, float maxStep, vec3 step, vec3 pos)
	{
		vec4 color = bg;
		vec3 uv;
		bool found = false;
		for (int i = 0; i < maxStep; ++i) {
			uv = pos / vec3(1,volumeAspect) + vec3(0.5, 0.5, 0.5);
			float density = getDensity_restricted(textureData, uv, densityScale);

			// simple iso
			if (density > isoValue) {
				//color = vec4(1, 0, 0, 1);
				found = true;
				break;
			}

			pos += step;
		}
		if (found)
		{
			float k = (1.0 - lightingStrength) + lightingStrength * dot(normalize(oLightDir.xyz), getNormal(textureData, uv));
			color = vec4(k * lightColor.xyz, 1);
		}
		return color;
	}


	void main() {
		vec2 uv = 2*fp.xy - vec2(1,1); // [0,1] to [-1,1]
		uv.y *= aspect;

		Ray ray;
		ray.origin = invViewMat[3].xyz;
		vec4 oDir = normalize(vec4(uv, -2, 0));
		vec4 wDir = invViewMat*oDir;
		ray.direction = wDir.xyz;

		AABB aabb;
		aabb.boxMin = vec3(-0.5, -0.5, -0.5);
		aabb.boxMax = vec3( 0.5,  0.5,  0.5);
		aabb.boxMin.yz *= volumeAspect.xy;
		aabb.boxMax.yz *= volumeAspect.xy;
		float tnear, tfar;
		bool hit = intersectBox(ray, aabb, tnear, tfar);

		if (!hit)
		{
			outColor = vec4(bgColor.xyz,1);
			return;
		}

		tnear = max(0.0, tnear); // handle the case when we are inside the volume
		float maxStep = max(volumeResolution.x, max(volumeResolution.y, volumeResolution.z)) * quality;
		vec3 step = ray.direction * (tfar - tnear) / maxStep;
		vec3 pos = ray.origin + (tnear + 0.0001) * ray.direction;

		vec4 sum = vec4(bgColor.xyz,0);
		
		switch(renderMode)
		{
		case 0:
			outColor = alphaBlend(sum, maxStep, step, pos);
			break;
		case 1:
			outColor = xray(sum, maxStep, step, pos);
			break;
		case 2:
			outColor = mip(sum, maxStep, step, pos);
			break;
		case 3:
			outColor = isoSurface(sum, maxStep, step, pos);
			break;

		}

	}
)";