#pragma once

#include <memory>
#include <filesystem>

#ifdef __cpp_lib_filesystem
	namespace fs = std::filesystem;
#elif __cpp_lib_experimental_filesystem
	namespace fs = std::experimental::filesystem;
#elif _MSC_VER <= 1900 && _MSC_VER > 1800
	namespace fs = std::experimental::filesystem;
#else
	#error "No filesystem support!"
#endif

#include "filtering/Volume.h"

namespace PETFilter
{

	class DataSet
	{
	public:
		bool LoadDat(const fs::path& path);
		bool LoadDicom(const fs::path& path);
		bool LoadDicomFiles(const fs::path& path);
		bool LoadDynamicDicomFiles(const fs::path& path, int frameIdx);
		bool LoadDynamicDicomFiles(const fs::path& path);
		bool LoadRaw(const fs::path& path, const int width, const int height, const int depth);
		bool LoadTxt(const fs::path& path, const int width, const int height, const int depth);

		void SetVolume(const std::shared_ptr<Volume>& v) { mVolumes[iVolume] = v; }
		std::shared_ptr<Volume> GetVolume() { return mVolumes[iVolume]; };
		void SelectVolume(int i = 0) { iVolume = i; }
		int  GetSelectedVolumeIdx( ) { return iVolume; }
		void GetAllVolumes(std::shared_ptr<Volume> volumes[]) { 
			for(int i = 0; i < nVolumes; i++) volumes[i] = mVolumes[i]; 
		}
		const fs::path& GetRepresentativePath() const {
			return mRepresentativeFilePath;
		}
	private:
		std::shared_ptr<Volume> mVolumes[nVolumes];
		int iVolume = 0;
		fs::path mRepresentativeFilePath;
	};	

}
