#include "image.h"

#include <FreeImage.h>
#include "dicom.h"

namespace PETFilter
{

bool saveImage(const std::wstring& fileName, const int width, const int height, const std::shared_ptr<std::vector<uint16_t>> data)
{
	size_t bpp = 16;

	FIBITMAP* bitmap = FreeImage_AllocateT(FIT_UINT16, (int)width, (int)height, (int)bpp);
	size_t pitch = FreeImage_GetPitch(bitmap);
	if (bitmap)
	{
		unsigned char* bits = FreeImage_GetBits(bitmap);
		if (pitch != width * 2)
		{
			for (size_t i = 0; i < height; ++i)
			{
				//memcpy(bits + (i * pitch), img->get<unsigned char>() + i * width * Image::elemSize(targetType) * Image::channels(targetType), width * (bpp / (size_t)8));
			}
		}
		else
		{
			memcpy(bits, data->data(), width * height * (bpp / (size_t)8));
		}
		// TODO: Do we realy need to flip?
		// FreeImage_FlipVertical(bitmap);
		FreeImage_SaveU(FIF_PNG, bitmap, fileName.c_str());
	}
	return true;
}

bool saveImage(const std::string& fileName, const std::string& refFileName, const int width, const int height, const std::shared_ptr<std::vector<uint16_t>> data)
{
	saveDicomImage(fileName, width, height, 1, *data.get(), refFileName);
	return true;
}

bool saveImage(const std::string& fileName, const int width, const int height, const std::shared_ptr<std::vector<uint16_t>> data, double(&origin)[3], double(&spacing)[3], float& slope, float& intercept)
{
	DicomImageDesc desc;
	desc.width = width;
	desc.height = height;
	desc.origin[0] = origin[0];
	desc.origin[1] = origin[1];
	desc.origin[2] = origin[2];
	desc.spacing[0] = spacing[0];
	desc.spacing[1] = spacing[1];
	desc.spacing[2] = spacing[2];

	saveDicomImage(fileName, *data.get(), desc);
	return true;
}

}

