#include "raw.h"
#include "filtering/commonDefines.h"
#include <fstream>
#include <sstream>

void loadFromVox(const std::string& filename, Volume& volume)
{
	std::ifstream inFile(filename, std::ios::in | std::ios::binary);
	VERIFY_CONDITION(inFile.good(), "Volume " + filename + " is not found");
	inFile.read(reinterpret_cast<char*>(volume.getHostPtr()), volume.getWidth() * volume.getHeight() * volume.getDepth() * sizeof(float));
}

void writeToVox(const std::string& filename, Volume& volume, bool appendDimensionsToFilename /* = true */)
{
	std::string fullFilename = filename;
	if (appendDimensionsToFilename)
	{
		std::stringstream sstr;
		sstr << '_' << volume.getWidth() << 'x' << volume.getHeight() << 'x' << volume.getDepth();
		fullFilename += sstr.str();
	}
	fullFilename += ".vox";

	std::ofstream outFile(fullFilename, std::ios_base::out | std::ios_base::binary);
	VERIFY_CONDITION(outFile.good(), "Unable to open " + fullFilename + " for writing");
	outFile.write(reinterpret_cast<const char*>(volume.getHostPtr()), volume.getWidth() * volume.getHeight() * volume.getDepth() * sizeof(float));
}

void loadFromTxt(const std::string& filename, Volume& volume)
{
	std::ifstream inFile(filename, std::ios::in);
	VERIFY_CONDITION(inFile.good(), "Volume " + filename + " is not found");
	float value;
	int idx = 0;
	while (!inFile.eof()) {
		inFile >> value;
		volume[idx++] = value;
	}
	inFile.close();
}

int numberOfItems(const std::string& filename)
{
	std::ifstream inFile(filename, std::ios::in);
	VERIFY_CONDITION(inFile.good(), "Volume " + filename + " is not found");
	float value;
	int idx = 0;
	while (!inFile.eof()) {
		inFile >> value;
		idx++;
	}
	inFile.close();
	return idx - 1;
}
